/*======================= PCYACC:  ======================================
           ABRAXAS SOFTWARE (R) PC YACC
       PCYACC (C) 1986-1991, ABRAXAS SOFTWARE, INC.
               all rights reserved
========================================================================*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "param.h"

/* ============================================================================ */
/* Function prototype declarations.                                             */
/* ============================================================================ */
extern int GetToken();
int        DescriptionSection();
int        GrammarRuleSection();
int        AppendInvalidItem();
int        InvalidItemInTable();
int        TokenInTable();
int        AppendNonterminal();
int        NonterminalInTable();
int        Product();
int        RuleSet();
int        LefthandSide();
int        RighthandSide();
void       Init();
void       ProgramSection();
void       SetupGrammarTree();
void       Rules();
void       AppendNextNode();
void       AppendToken();
void       AppendRightNode();
void       RuleWithIndex1();
void       RuleWithIndex2();
void       RuleWithIndex3();
void       CRTableInOrdinaryOrder();
void       Sorting1();
void       Sorting2();
void       CRTableInAlpha();
void       CRTableInRuleNumber();
void       CreatCrossReferenceTable();
void       SourceCodeWithLineNumber();
void       ReleaseMemory();
void       RuleWithoutIndex();

#define MAX_TOKEN          500   /* Define the default number of tokens in      */
				 /* a grammar description file.                 */
#define MAX_NONTERMINAL  1000    /* Define the default number of nonterminals   */
				 /* in a grammar description file.              */
#define MAX_INVALID        200   /* Define the default number of invalid items  */
				 /* in a grammar description file.              */

#define YYDEBUG

/* The data structure for nonterminal array node.                               */
struct NONTERMINAL_NODE {
			     long LineNumber;
			     int  RuleNumber;
			     char *Text;
			 };

struct RULE_NODE {
		       struct RULE_NODE *Next;
		       int    ItemNumber;
		       int    Attribute;
		       struct RULE_NODE *Right;
		 } ;


struct NONTERMINAL_NODE NonTerminalTable[MAX_NONTERMINAL];
struct RULE_NODE  *RuleTreeRoot;
struct RULE_NODE  *LastTreeNode;
struct RULE_NODE  *CurrentTreeNode;
struct RULE_NODE  *RightSubtreeRoot;
struct RULE_NODE  *NextSubtreeRoot;

char   *TokenTable[MAX_TOKEN];         /* The table for holding tokens. This an */
				       /* array of char pointer. The space it   */
				       /* points is alocated dynamically for    */
				       /* hold the content of a certain token.  */
char   *InvalidItemTable[MAX_INVALID]; /* The table for holding invalid items.  */
				       /* This is an array of char pointer,the  */
				       /* space it  points is alocated dynami-  */
				       /* cally for holding certain content of  */
				       /* an invalid item.                      */
char   TokenBuffer[TOKEN_LENGTH];      /* The buffer for holding the content of */
				       /* current token.                        */
char   LastBuffer[TOKEN_LENGTH];       /* The buffer for holding the content of */
				       /* the last token when looking forward to*/
				       /* the context.                          */

long LineNumber         = 1;           /* The counter for counting the lines of */
				       /* YACC source code.                     */
int  NewRule            = 0;           /* A new rule begin flag.                */
int  LookAhead          = 0;           /* Lookahead flag.                       */
int  AheadToken         = 0;           /* The buffer for holding the ahead look-*/
				       /* ing token.                            */
int  NextBegin          = 0;           /* The next rule of the same left part   */
				       /* begining flag.                        */
int  TokenNumber        = -1;
int  NonterminalNumber  = -1;
int  InvalidItemNumber  = -1;
int  RuleNumber         =  1;
char *Msg[]={
    " Usage:YTOOL <SourceFileName> <Options>",
    "       <SourceFileName>: Grammar Description File Name",
    "       <Options>       : a | b | c | d | e | f | g |h ",
    "       a. Create GRAMMAR FOREST from YACC source file",
    "       b. Create GRAMMAR FOREST with rule & line number (1) ",
    "       c. Create GRAMMAR FOREST with rule & line number (2) ",
    "       d. Create GRAMMAR FOREST with rule & line number (3) ",
    "       e. Cross Reference Table of the rule in the rule appearing order",
    "       f. Cross Reference Table of the rule in the order of alpha",
    "       g. Cross Reference Table of the rule in nonterminal appearing order",
    "       h. Grammar Description File with line number",
    NULL,
};


FILE *SFile;

void
Init()
{
   RuleTreeRoot    = (struct RULE_NODE *) malloc(sizeof(struct RULE_NODE));
   if (RuleTreeRoot==NULL)
   {
	printf("No enough memory !\n");
	exit(-1);
   }
   RuleTreeRoot->Right = (struct RULE_NODE *) NULL;
   RuleTreeRoot->Next  = (struct RULE_NODE *) NULL;
   NextSubtreeRoot     = RuleTreeRoot;
}


int
DescriptionSection()
{
     int NextToken = 0;

     while ((NextToken=GetToken())!=PERCENT_RBRACE)
     {
	    if (NextToken==EOF)
	    {
		fprintf(stderr,"Illigle Input,Abort!\n");
		exit(-1);
	    }
     }
     NextToken=GetToken();
     return NextToken;
}

int
AppendInvalidItem(Item)
char *Item;
{
   char *Text;

   if(++InvalidItemNumber>=MAX_INVALID)
   {
       fprintf(stderr,"Invalid item table overflow!\n");
       exit(-1);
   }
   Text=(char *) malloc(strlen(Item)+1);
   if (Text==NULL)
      {
	   printf("No enough memory \n");
	   exit (-1);
      }
   strcpy(Text,Item);
   InvalidItemTable[InvalidItemNumber]=Text;
   return InvalidItemNumber;
}

int
InvalidItemInTable(Text)
char *Text;
{
  int i=0;
  int ReturnValue=(-1);

  if (InvalidItemNumber>=0)
  {
      while (i<=InvalidItemNumber&&(strcmp(InvalidItemTable[i],Text)!=0))
	 i++;
      if (i<=InvalidItemNumber)
      if ((strcmp(InvalidItemTable[i],Text))==0)
	  ReturnValue=i;
  }
  return ReturnValue;
}


void
AppendToken(buf)
char *buf;
{
   char *Text;

   if(++TokenNumber>=MAX_TOKEN)
   {
      fprintf(stderr,"Token table overflow\n");
      exit(-1);
   }
   Text=(char *) malloc(strlen(buf)+1);
   if (Text==NULL)
   {
	printf("No enough memory \n");
	exit (-1);
   }
   strcpy(Text,buf);
   TokenTable[TokenNumber]=Text;
}

int
TokenInTable(Text)
char *Text;
{
  int i=0;
  int ReturnValue=-1;

  if (TokenNumber>=0)
  {
      while (i<=TokenNumber && (strcmp(TokenTable[i],Text)!=0))
	 i++;
      if (i<=TokenNumber)
      if ((strcmp(TokenTable[i],Text))==0)
	  ReturnValue=i;
  }
  return ReturnValue;
}

int
AppendNonterminal(buf)
char *buf;
{
   char *Text;

   if(++NonterminalNumber>=MAX_NONTERMINAL)
   {
      fprintf(stderr,"Nonterminater table overflow !\n");
      exit(-1);
   }
   Text=(char *) malloc(strlen(buf)+1);
   if (Text==NULL)
   {
      printf("No enough memory \n");
      exit (-1);
   }
   strcpy(Text,buf);
   NonTerminalTable[NonterminalNumber].Text=Text;
   return NonterminalNumber;
}

int
NonterminalInTable(Text)
char *Text;
{
  int i=0;
  int ReturnValue=-1;

  if (NonterminalNumber>=0)
  {
    while (i<=NonterminalNumber && (strcmp(NonTerminalTable[i].Text,Text)!=0))
	 i++;
    if (i<=NonterminalNumber)
    if ((strcmp(NonTerminalTable[i].Text,Text))==0)
	  ReturnValue=i;
  }
  return ReturnValue;
}

int
GrammarRuleSection()
{
   int Token   = 0;
   int ReturnValue;
   while(1)
   {
       Token=GetToken();
       if (Token==TOKEN||Token==PREC||Token==LEFT||Token==RIGHT||Token==NONASSOC)
       {
	  Token=GetToken();
	  do
	  {
	      /* Store the Tokens */
	      if (Token==IDENTIFIER)
	      {
		if ((ReturnValue=TokenInTable(TokenBuffer))==-1)
		  AppendToken(TokenBuffer);
	      } else {
			if ((ReturnValue=InvalidItemInTable(LastBuffer))==-1)
			    ReturnValue= AppendInvalidItem(LastBuffer);
		     }
	      Token=GetToken();
	  } while ((Token!=PERCENT) && (Token!=PERCENT_PERCENT));
	}
	else do
		   Token=GetToken();
	     while ((Token!=PERCENT) && (Token!=PERCENT_PERCENT));
       if (Token==EOF)
	   break;
       if (Token==PERCENT_PERCENT)
	   break;
   }
   return Token;
}

void
ProgramSection()
{
    int NextToken;
    NextToken=GetToken();
    if (NextToken==EOF)
      {
	 fprintf(stderr,"Illigle Input,Abort!\n");
	 exit(-1);
      }
    Rules(NextToken);
    /* return value processing */
}

void
SetupGrammarTree()
{
   int Token      = 0;
 /*  char *content;  */

   Token=GetToken();
   while (Token!=PERCENT_PERCENT && Token!=EOF)
   {
     if (Token==EOF)
	 {
	    fprintf(stderr,"Unexpected end of file \n");
	    exit (-1);
	 }
     while (Token!=PERCENT_LBRACE && Token!=PERCENT && Token!=PERCENT && Token !=EOF)
	 Token= GetToken();
     if (Token== PERCENT_LBRACE)
	 Token=DescriptionSection();
     if (Token==PERCENT)
	 Token=GrammarRuleSection();
   }
   if (Token== PERCENT_PERCENT)
	 ProgramSection();
}

void
Rules(Token)
int Token;
{
     int LocalToken=Token;
     if (Token==EOF)
      {
	 fprintf(stderr,"Unexpected end of file\n");
	 exit(-1);
      }
     LocalToken=Product(Token);
     while (LocalToken!=IDENTIFIER
		 && LocalToken!=STRING
		 && LocalToken!=PERCENT
		 && LocalToken!=PREC
		 && LocalToken!=PERCENT_PERCENT
		 && LocalToken!=EOF)
	   LocalToken=GetToken();
     if (LocalToken==IDENTIFIER
	      || LocalToken==STRING
	      || LocalToken==PREC
	      || LocalToken==PERCENT)
	   Rules(LocalToken);
}

int
Product(Token)
int Token;
{
   int LocalToken;

   LocalToken=LefthandSide(Token);
   if (LocalToken!=COLON)
      fprintf(stderr,"WARNING Line %d ':' is expected\n",LineNumber);
   RuleNumber++;
   LocalToken=RighthandSide(GetToken());
   if (LocalToken!=SEMICOLON)
	   fprintf(stderr,"WARNING before Line %d ';' is expected\n",LineNumber);
	else
	   LocalToken=GetToken();
   return LocalToken;
}

void
AppendNextNode(ItemNumber,Attribute)
int ItemNumber;
int Attribute;
{

    if ((CurrentTreeNode = (struct RULE_NODE *)
		malloc(sizeof(struct RULE_NODE)))==NULL)
      {
	  printf("No enough memory !\n");
	  exit(-1);
      }
    CurrentTreeNode->Right = (struct RULE_NODE *) NULL;
    CurrentTreeNode->Next  = (struct RULE_NODE *) NULL;
    CurrentTreeNode->ItemNumber = ItemNumber;
    NextSubtreeRoot->Next  = CurrentTreeNode;
    NextSubtreeRoot        = CurrentTreeNode;
    LastTreeNode           = CurrentTreeNode;
    CurrentTreeNode -> Attribute = Attribute ;
    NewRule=1;
    NextBegin=0;
}

int
LefthandSide(Token)
int Token;
{
  int  LocalToken=Token;
  int  ItemNumber;
  int  Attribute;

  Attribute =1;
  if ((ItemNumber=TokenInTable(TokenBuffer))!=-1)
     {
       fprintf(stderr,
	    "TERMINATER CAN NOT appear in the left of a Product!,Line %d\n",LineNumber);
       Attribute=0;
     } else if (LocalToken!=IDENTIFIER)
	 {
	   fprintf(stderr,"ERROR Line %d IDENTIFIER is expected \n",LineNumber);
	   if ((ItemNumber=InvalidItemInTable(TokenBuffer))==-1)
	       ItemNumber=AppendInvalidItem(TokenBuffer);
	   Attribute = -1;
	 }
       else {
	      if ((ItemNumber=NonterminalInTable(TokenBuffer))==-1)
		  ItemNumber=AppendNonterminal(TokenBuffer);
	      NonTerminalTable[ItemNumber].LineNumber=LineNumber;
	      NonTerminalTable[ItemNumber].RuleNumber=RuleNumber;
	      Attribute=2;
	    }

   AppendNextNode(ItemNumber,Attribute);
  if (LookAhead)
    {
      LookAhead=0;
      LocalToken=AheadToken;
    }
  else
     if ((LocalToken=GetToken())==EOF)
     {
      fprintf(stderr,"Unexpected end of file \n");
      exit (-1);
     }
  return LocalToken;
}

void
AppendRightNode(ItemNumber,Attribute)
int ItemNumber;
int Attribute;
{
	 CurrentTreeNode = (struct RULE_NODE *)
				malloc(sizeof(struct RULE_NODE));
	 if (CurrentTreeNode==(struct RULE_NODE *) NULL)
	  {
	     printf("No enough memory \n");
	     exit (-1);
	  }
	  CurrentTreeNode->Right = (struct RULE_NODE *) NULL;
	  CurrentTreeNode->Next  = (struct RULE_NODE *) NULL;
	  CurrentTreeNode->ItemNumber = ItemNumber;
	  CurrentTreeNode ->Attribute = Attribute ;
	  if (NextBegin)
	     {
		RightSubtreeRoot->Next = CurrentTreeNode;
		RightSubtreeRoot       = CurrentTreeNode;
		LastTreeNode           = CurrentTreeNode;
		NextBegin =0;
	    } else  {
			if (NewRule)
			{
			     RightSubtreeRoot       = CurrentTreeNode;
			     NewRule = 0;
			 }
		      LastTreeNode->Right    = CurrentTreeNode;
		      LastTreeNode           = CurrentTreeNode;
		    }
}

int
RuleSet(Token)
int Token;
{
   int LocalToken=Token;
   int ItemNumber;
   int Attribute;

   if (LocalToken==SEMICOLON)
      {
	 /* may be empty */
	 if (NewRule)
	       AppendRightNode(0,0);
	 return LocalToken;
      }
   if (LocalToken==VLINE)
   {
      /* empty */
      AppendRightNode(0,0);
      NextBegin = 1;
      return LocalToken;
   } else
	   do {
		strcpy(LastBuffer,TokenBuffer);
		AheadToken=GetToken();
		if (AheadToken==COLON)
		  {
		     LookAhead=1;
		     strcpy(TokenBuffer,LastBuffer);
		     break;
		  } else if (AheadToken!=IDENTIFIER
			      && AheadToken!=STRING
			      && AheadToken!=PREC
			      && AheadToken!=PERCENT)
			    {
				  if (LocalToken==IDENTIFIER)
				   {
				     Attribute = 1;
				     if ((ItemNumber=TokenInTable(LastBuffer))==-1)
				     {
					Attribute=2;
					if ((ItemNumber=NonterminalInTable(LastBuffer))==-1)
					   ItemNumber= AppendNonterminal(LastBuffer);
				     }
				   } else {
					      Attribute=-1;
					      if ((ItemNumber=
						   InvalidItemInTable(LastBuffer))==-1)
						ItemNumber= AppendInvalidItem(LastBuffer);
					    }

				AppendRightNode(ItemNumber,Attribute);

				LocalToken=AheadToken;
				break;
			    }
		   if (LocalToken==IDENTIFIER)
		    {
		       Attribute = 1;
		       if ((ItemNumber=TokenInTable(LastBuffer))==-1)
		       {
			Attribute=2;
			if ((ItemNumber=NonterminalInTable(LastBuffer))==-1)
			   ItemNumber=AppendNonterminal(LastBuffer);
		      }
		   } else {
			       Attribute=-1;
			       if ((ItemNumber=InvalidItemInTable(LastBuffer))==-1)
				    ItemNumber= AppendInvalidItem(LastBuffer);
			     }

		AppendRightNode(ItemNumber,Attribute);
		LocalToken=AheadToken;
	   } while (1);
   return LocalToken;
}

int
RighthandSide(Token)
int Token;
{
  int LocalToken;

  LocalToken=RuleSet(Token);
  if (LocalToken==VLINE)
     {
	RuleNumber++;
	NextBegin=1;
	LocalToken=GetToken();
	LocalToken=RighthandSide(LocalToken);
     }
  else
       return LocalToken;
}

void
RuleWithoutIndex()
{
  int First=1;
  int Index  ;
  int Attribute;

  printf("--------------------------------------------------------------------------------\n");
  printf("                             Grammar Forest\n");
  printf("--------------------------------------------------------------------------------\n");

  NextSubtreeRoot=RuleTreeRoot;
  while (NextSubtreeRoot->Next!=NULL)
  {
      NextSubtreeRoot=NextSubtreeRoot->Next;
      Attribute=NextSubtreeRoot->Attribute;
      Index    =NextSubtreeRoot->ItemNumber;
      switch(Attribute)
      {
	    case -1:printf("%-16s: ",InvalidItemTable[Index]);
		    break;
	    case  0:printf("    : ");
		    break;
	    case  1:printf("%-16s: ",TokenTable[Index]);
		    break;
	    case  2:
		    printf("%-16s: ",NonTerminalTable[Index].Text);
		    break;
      }
      First=1;
      RightSubtreeRoot=NextSubtreeRoot->Right;
      while (RightSubtreeRoot!=NULL)
      {
	  if (First)
	      First=0;
	  else  printf("\n\t\t| ");
	  LastTreeNode=RightSubtreeRoot;
	  while(LastTreeNode)
	  {
	      Attribute=LastTreeNode->Attribute;
	      Index    =LastTreeNode->ItemNumber;
	      switch(Attribute)
		{
		    case -1:printf("%s ",InvalidItemTable[Index]);
			    break;
		    case  0:printf("    ");
			    break;
		    case  1:printf("%s ",TokenTable[Index]);
			    break;
		    case  2:printf("%s ",NonTerminalTable[Index].Text);
			    break;
	      }
	      LastTreeNode=LastTreeNode->Right;
	  }
	  RightSubtreeRoot=RightSubtreeRoot->Next;
      }
      printf("\n;\n");

  }
}

void
RuleWithIndex1()
{
  int First=1;
  int Index  ;
  int Attribute;

  printf("--------------------------------------------------------------------------------\n");
  printf("              Grammar Forest With Rule Number & Line Number (FORM 1)\n");
  printf("--------------------------------------------------------------------------------\n");
  NextSubtreeRoot=RuleTreeRoot;
  while (NextSubtreeRoot->Next!=NULL)
  {
      NextSubtreeRoot=NextSubtreeRoot->Next;
      Attribute=NextSubtreeRoot->Attribute;
      Index    =NextSubtreeRoot->ItemNumber;
      switch(Attribute)
      {
	    case -1:printf("%-16s: ",InvalidItemTable[Index]);
		    break;
	    case  0:printf("    : ");
		    break;
	    case  1:printf("-16%s: ",TokenTable[Index]);
		    break;
	    case  2:
		    printf("[%-3d,%-4d]\n",NonTerminalTable[Index].RuleNumber,
					   NonTerminalTable[Index].LineNumber);
		    printf("%-16s: ",NonTerminalTable[Index].Text);
		    break;
      }
      First=1;
      RightSubtreeRoot=NextSubtreeRoot->Right;
      while (RightSubtreeRoot!=NULL)
      {
	  if (First)
	      First=0;
	  else  printf("\n\t\t| ");
	  LastTreeNode=RightSubtreeRoot;
	  while(LastTreeNode)
	  {
	      Attribute=LastTreeNode->Attribute;
	      Index    =LastTreeNode->ItemNumber;
	      switch(Attribute)
		{
		    case -1:printf("%s ",InvalidItemTable[Index]);
			    break;
		    case  0:printf("    ");
			    break;
		    case  1:printf("%s ",TokenTable[Index]);
			    break;
		    case  2:printf("%s ",NonTerminalTable[Index].Text);
			    break;
	      }
	      LastTreeNode=LastTreeNode->Right;
	  }
	  RightSubtreeRoot=RightSubtreeRoot->Next;
      }
      printf("\n;\n");

  }
}

void
RuleWithIndex2()
{
  int First=1  ;
  int Index    ;
  int Attribute;
  int rules    ;

  printf("--------------------------------------------------------------------------------\n");
  printf("              Grammar Forest With Rule Number & Line Number (FORM 2)\n");
  printf("--------------------------------------------------------------------------------\n");
  NextSubtreeRoot=RuleTreeRoot;
  while (NextSubtreeRoot->Next!=NULL)
  {
      NextSubtreeRoot=NextSubtreeRoot->Next;
      Attribute=NextSubtreeRoot->Attribute;
      Index    =NextSubtreeRoot->ItemNumber;
      switch(Attribute)
      {
	    case -1:printf("%s: ",InvalidItemTable[Index]);
		    break;
	    case  0:printf("    :");
		    break;
	    case  1:printf("%s :",TokenTable[Index]);
		    break;
	    case  2:
		    rules = NonTerminalTable[Index].RuleNumber;
		    printf("\n[%-3d,%-4d]\n",rules++,NonTerminalTable[Index].LineNumber);
		    printf("%-16s: ",NonTerminalTable[Index].Text);
		    break;
      }
      First=1;
      RightSubtreeRoot=NextSubtreeRoot->Right;
      while (RightSubtreeRoot!=NULL)
      {
	  if (First)
	      First=0;
	  else  printf("\n[%-3d]\t\t| ",rules++);
	  LastTreeNode=RightSubtreeRoot;
	  while(LastTreeNode)
	  {
	      Attribute=LastTreeNode->Attribute;
	      Index    =LastTreeNode->ItemNumber;
	      switch(Attribute)
		{
		    case -1:printf("%s ",InvalidItemTable[Index]);
			    break;
		    case  0:printf("    ");
			    break;
		    case  1:printf("%s ",TokenTable[Index]);
			    break;
		    case  2:printf("%s ",NonTerminalTable[Index].Text);
			    break;
	      }
	      LastTreeNode=LastTreeNode->Right;
	  }
	  RightSubtreeRoot=RightSubtreeRoot->Next;
      }
      printf("\n;\n");

  }
}

void
RuleWithIndex3()
{
  int First=1  ;
  int Index    ;
  int Attribute;
  int rules    ;

  printf("--------------------------------------------------------------------------------\n");
  printf("              Grammar Forest With Rule Number & Line Number (FORM 3)\n");
  printf("--------------------------------------------------------------------------------\n");
  NextSubtreeRoot=RuleTreeRoot;
  while (NextSubtreeRoot->Next!=NULL)
  {
      NextSubtreeRoot=NextSubtreeRoot->Next;
      Attribute=NextSubtreeRoot->Attribute;
      Index    =NextSubtreeRoot->ItemNumber;
      switch(Attribute)
      {
	    case -1:printf("%s: ",InvalidItemTable[Index]);
		    break;
	    case  0:printf("    :");
		    break;
	    case  1:printf("%s: ",TokenTable[Index]);
		    break;
	    case  2:
		    rules = NonTerminalTable[Index].RuleNumber;
		    printf("\n[%-3d,%-4d]\n",rules++,NonTerminalTable[Index].LineNumber);
		    printf("%-16s: ",NonTerminalTable[Index].Text);
		    break;
      }
      First=1;
      RightSubtreeRoot=NextSubtreeRoot->Right;
      while (RightSubtreeRoot!=NULL)
      {
	  if (First)
	      First=0;
	  else  printf("\n[%-3d]\t\t| ",rules++);
	  LastTreeNode=RightSubtreeRoot;
	  while(LastTreeNode)
	  {
	      Attribute=LastTreeNode->Attribute;
	      Index    =LastTreeNode->ItemNumber;
	      switch(Attribute)
		{
		    case -1:printf("%s ",InvalidItemTable[Index]);
			    break;
		    case  0:printf("    ");
			    break;
		    case  1:printf("%s ",TokenTable[Index]);
			    break;
		    case  2:printf("%s[*%d] ",NonTerminalTable[Index].Text,
					   NonTerminalTable[Index].RuleNumber);
			    break;
	      }
	      LastTreeNode=LastTreeNode->Right;
	  }
	  RightSubtreeRoot=RightSubtreeRoot->Next;
      }
      printf("\n;\n");

  }
}

void
CRTableInOrdinaryOrder()
{
  int  i;

  printf("   ===========  CROSS REFERENCE TABLE (1) ===========\n");
  printf("\t RULE_No.| LINE_No.| SYMBOL \n");
  printf("   --------------------------------------------------\n");
  for (i=0;i<=NonterminalNumber;i++)
    {
       if (NonTerminalTable[i].RuleNumber)
       {
	   printf("\t %-8d|",NonTerminalTable[i].RuleNumber);
	   printf(" %-8d|",NonTerminalTable[i].LineNumber);
	   printf(" %s\n",NonTerminalTable[i].Text);
       }
    }
  printf("   --------------------------------------------------\n");
}

void
Sorting1(l,r)
int l,r;
{
  int i;
  int j;
  int min;
  int tmp1,tmp2;
  char *tmp3,*minc;

  for (i=l;i<=r;i++)
  {
      min=i;
      minc=NonTerminalTable[i].Text;
      for (j=i;j<=r;j++)
      {
	 if (strcmp(NonTerminalTable[j].Text,minc)<0)
	   {
	      min=j;
	      minc=NonTerminalTable[j].Text;
	   }
      }
	  tmp1=NonTerminalTable[i].RuleNumber;
	  tmp2=NonTerminalTable[i].LineNumber;
	  tmp3=NonTerminalTable[i].Text;
	  NonTerminalTable[i  ].RuleNumber=NonTerminalTable[min].RuleNumber;
	  NonTerminalTable[i  ].LineNumber=NonTerminalTable[min].LineNumber;
	  NonTerminalTable[i  ].Text   =NonTerminalTable[min].Text;
	  NonTerminalTable[min].RuleNumber=tmp1;
	  NonTerminalTable[min].LineNumber=tmp2;
	  NonTerminalTable[min].Text   =tmp3;

  }
}

void
Sorting2(l,r)
int l,r;
{
  int i;
  int j;
  int min,minc;
  int tmp1,tmp2;
  char *tmp3;

  for (i=l;i<=r;i++)
  {
      min=i;
      minc=NonTerminalTable[i].RuleNumber;
      for (j=i;j<=r;j++)
      {
	 if (NonTerminalTable[j].RuleNumber<minc)
	   {
	      min=j;
	      minc=NonTerminalTable[j].RuleNumber;
	   }
      }
	  tmp1=NonTerminalTable[i].RuleNumber;
	  tmp2=NonTerminalTable[i].LineNumber;
	  tmp3=NonTerminalTable[i].Text;
	  NonTerminalTable[i  ].RuleNumber=NonTerminalTable[min].RuleNumber;
	  NonTerminalTable[i  ].LineNumber=NonTerminalTable[min].LineNumber;
	  NonTerminalTable[i  ].Text   =NonTerminalTable[min].Text;
	  NonTerminalTable[min].RuleNumber=tmp1;
	  NonTerminalTable[min].LineNumber=tmp2;
	  NonTerminalTable[min].Text   =tmp3;

  }
}

void
CRTableInAlpha()
{
  int i;
   /* Sorting */
  Sorting1(0,NonterminalNumber);
  printf("   ===========  CROSS REFERENCE TABLE (2) ===========\n");
  printf("\t RULE_No.| LINE_No.| SYMBOL \n");
  printf("   --------------------------------------------------\n");
  for (i=0;i<=NonterminalNumber;i++)
    {
       if (NonTerminalTable[i].RuleNumber)
       {
	   printf("\t %-8d|",NonTerminalTable[i].RuleNumber);
	   printf(" %-8d|",NonTerminalTable[i].LineNumber);
	   printf(" %s\n",NonTerminalTable[i].Text);
       }
    }
  printf("   --------------------------------------------------\n");
}

void
CRTableInRuleNumber()
{
  int i;
/*  int j;  */

  Sorting2(0,NonterminalNumber);
  printf("   ===========  CROSS REFERENCE TABLE (3) ===========\n");
  printf("\t RULE_No.| LINE_No.| SYMBOL \n");
  printf("   --------------------------------------------------\n");
  for (i=0;i<=NonterminalNumber;i++)
    {
       if (NonTerminalTable[i].RuleNumber)
       {
	   printf("\t %-8d|",NonTerminalTable[i].RuleNumber);
	   printf(" %-8d|",NonTerminalTable[i].LineNumber);
	   printf(" %s\n",NonTerminalTable[i].Text);
       }
    }
  printf("   --------------------------------------------------\n");
}

void
CreatCrossReferenceTable(Selection)
char Selection;
{
   switch(Selection)
   {
	case 'e':
	       CRTableInOrdinaryOrder();
	       break;
	case 'f':
	       CRTableInAlpha();
	       break;
	case 'g':
	       CRTableInRuleNumber();
	       break;
   }
}

void
SourceCodeWithLineNumber()
{
   char Buf[150];
   int  Line=1;

   fseek(SFile,0L,0);
   while (fgets(Buf,149,SFile)!=NULL)
      printf("%-4d: %s",Line++,Buf);
}

void
ReleaseMemory()
{
  int i;
  struct RULE_NODE  *ReleasedNode;

  /* Releasing the tree */
  NextSubtreeRoot=RuleTreeRoot;
  while (NextSubtreeRoot->Next!=NULL)
  {
      ReleasedNode    =NextSubtreeRoot;
      NextSubtreeRoot =NextSubtreeRoot->Next;
      RightSubtreeRoot=NextSubtreeRoot->Right;
      free(ReleasedNode);
      while (RightSubtreeRoot!=NULL)
      {
	  LastTreeNode    =RightSubtreeRoot;
	  RightSubtreeRoot=RightSubtreeRoot->Next;
	  while(LastTreeNode)
	  {
	      ReleasedNode = LastTreeNode;
	      LastTreeNode=LastTreeNode->Right;
	      free(ReleasedNode);
	  }
      }
  }
  for (i=0;i<=TokenNumber;i++)
     free(TokenTable[i]);
  for (i=0;i<=InvalidItemNumber;i++)
     free(InvalidItemTable[i]);
  for (i=0;i<=NonterminalNumber;i++)
     free(NonTerminalTable[i].Text);
}

int
main(argc,argv)
int   argc;
char *argv[];
{
  char Selection;
  int Processed = 0;
  int i         = 0;

  if (argc<3)
  {
     do {
	 fprintf(stderr,"%s\n",Msg[i]);
	 i++;
     }while (Msg[i]!=NULL);
     exit(-1);
  };
  if ((SFile=fopen(argv[1],"r"))==NULL)
  {
    fprintf(stderr,"Can not open source file \"%s\" \n",argv[1]);
    exit(-1);
  };
  Selection=argv[2][0];
  if (Selection<'a' || Selection>'h')
  {
    fprintf(stderr,"Option error\n");
    exit(-1);
  }
    switch(Selection)
    {
      case 0:
	     if (Processed)
		 ReleaseMemory();
	     break;
      case 'a':
	     if (!Processed)
	     {
		 Init();
		 SetupGrammarTree();
		 Processed=1;
	     }
	     RuleWithoutIndex();
	     break;
      case 'b':
	     if (!Processed)
	     {
		 Init();
		 SetupGrammarTree();
		 Processed=1;
	     }
	     RuleWithIndex1();
	     break;
      case 'c':
	     if (!Processed)
	     {
		 Init();
		 SetupGrammarTree();
		 Processed=1;
	     }
	     RuleWithIndex2();
	     break;
      case 'd':
	     if (!Processed)
	     {
		 Init();
		 SetupGrammarTree();
		 Processed=1;
	     }
	     RuleWithIndex3();
	     break;
      case 'e':
      case 'f':
      case 'g':
	     if (!Processed)
	     {
		 Init();
		 SetupGrammarTree();
		 Processed=1;
	     }
	     CreatCrossReferenceTable(Selection);
	     break;
      case 'h':
	     SourceCodeWithLineNumber();
	     break;
    }
  return (0);
}



